from __future__ import annotations

from typing import TYPE_CHECKING

from viktor import AI  # module part of autodoc_mock_imports # noqa: F401

if TYPE_CHECKING:
    # Nested type guard
    from typing import Literal  # noqa: F401


class AnotherClass:
    """Another class is here"""
