def greet(name: str, greeting: str) -> str: ...

class Calculator:
    value: int
    def add(self, x: int) -> Calculator: ...
    class Inner:
        def process(self, data: bytes) -> bytes: ...

async def fetch(url: str) -> str: ...
